
import datetime
import docker
import hashlib
import json
import logging
import os
import re
import shutil
import six
import tarfile
import tempfile
import threading

from docker_squash.errors import SquashError, SquashUnnecessaryError

if not six.PY3:
    import docker_squash.lib.xtarfile


class Chdir(object):

    """ Context manager for changing the current working directory """

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


class Image(object):
    """
    Base class for all Docker image formats. Contains many functions that are handy
    while squashing the image.

    This class should not be used directly.
    """

    FORMAT = None
    """ Image format version """

    def __init__(self, log, docker, image, from_layer, tmp_dir=None, tag=None, comment=""):
        self.log = log
        self.debug = self.log.isEnabledFor(logging.DEBUG)
        self.docker = docker
        self.image = image
        self.from_layer = from_layer
        self.tag = tag
        self.comment = comment
        self.image_name = None
        self.image_tag = None
        self.squash_id = None

        # Workaround for https://play.golang.org/p/sCsWMXYxqy
        #
        # Golang doesn't add padding to microseconds when marshaling
        # microseconds in date into JSON. Python does.
        # We need to produce same output as Docker's to not generate
        # different metadata. That's why we need to strip all zeros at the
        # end of the date string...
        self.date = re.sub(
            r'0*Z$', 'Z', datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ'))
        """ Date used in metadata, already formatted using the `%Y-%m-%dT%H:%M:%S.%fZ` format """

        self.tmp_dir = tmp_dir
        """ Main temporary directory to save all working files. This is the root directory for all other temporary files. """

    def squash(self):
        self._before_squashing()
        ret = self._squash()
        self._after_squashing()

        return ret

    def _squash(self):
        pass

    def cleanup(self):
        """ Cleanup the temporary directory """

        self.log.debug("Cleaning up %s temporary directory" % self.tmp_dir)
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def _initialize_directories(self):
        # Prepare temporary directory where all the work will be executed
        try:
            self.tmp_dir = self._prepare_tmp_directory(self.tmp_dir)
        except:
            raise SquashError("Preparing temporary directory failed")

        # Temporary location on the disk of the old, unpacked *image*
        self.old_image_dir = os.path.join(self.tmp_dir, "old")
        # Temporary location on the disk of the new, unpacked, squashed *image*
        self.new_image_dir = os.path.join(self.tmp_dir, "new")
        # Temporary location on the disk of the squashed *layer*
        self.squashed_dir = os.path.join(self.new_image_dir, "squashed")

        for d in self.old_image_dir, self.new_image_dir:
            os.makedirs(d)

    def _squash_id(self, layer):
        if layer == "<missing>":
            self.log.warning(
                "You try to squash from layer that does not have it's own ID, we'll try to find it later")
            return None

        try:
            squash_id = self.docker.inspect_image(layer)['Id']
        except:
            raise SquashError(
                "Could not get the layer ID to squash, please check provided 'layer' argument: %s" % layer)

        if squash_id not in self.old_image_layers:
            raise SquashError(
                "Couldn't find the provided layer (%s) in the %s image" % (layer, self.image))

        self.log.debug("Layer ID to squash from: %s" % squash_id)

        return squash_id

    def _validate_number_of_layers(self, number_of_layers):
        """
        Makes sure that the specified number of layers to squash
        is a valid number
        """

        # Only positive numbers are correct
        if number_of_layers <= 0:
            raise SquashError(
                "Number of layers to squash cannot be less or equal 0, provided: %s" % number_of_layers)

        # Do not squash if provided number of layer to squash is bigger
        # than number of actual layers in the image
        if number_of_layers > len(self.old_image_layers):
            raise SquashError(
                "Cannot squash %s layers, the %s image contains only %s layers" % (number_of_layers, self.image, len(self.old_image_layers)))

    def _before_squashing(self):
        self._initialize_directories()

        # Location of the tar archive with squashed layers
        self.squashed_tar = os.path.join(self.squashed_dir, "layer.tar")

        if self.tag:
            self.image_name, self.image_tag = self._parse_image_name(self.tag)

        # The image id or name of the image to be squashed
        try:
            self.old_image_id = self.docker.inspect_image(self.image)['Id']
        except SquashError:
            raise SquashError(
                "Could not get the image ID to squash, please check provided 'image' argument: %s" % self.image)

        self.old_image_layers = []

        # Read all layers in the image
        self._read_layers(self.old_image_layers, self.old_image_id)

        self.old_image_layers.reverse()

        self.log.info("Old image has %s layers", len(self.old_image_layers))
        self.log.debug("Old layers: %s", self.old_image_layers)

        # By default - squash all layers.
        if self.from_layer == None:
            self.from_layer = len(self.old_image_layers)

        try:
            number_of_layers = int(self.from_layer)

            self.log.debug(
                "We detected number of layers as the argument to squash")
        except ValueError:
            self.log.debug("We detected layer as the argument to squash")

            squash_id = self._squash_id(self.from_layer)

            if not squash_id:
                raise SquashError(
                    "The %s layer could not be found in the %s image" % (self.from_layer, self.image))

            number_of_layers = len(self.old_image_layers) - \
                self.old_image_layers.index(squash_id) - 1

        self._validate_number_of_layers(number_of_layers)

        marker = len(self.old_image_layers) - number_of_layers

        self.layers_to_squash = self.old_image_layers[marker:]
        self.layers_to_move = self.old_image_layers[:marker]

        self.log.info("Checking if squashing is necessary...")

        if len(self.layers_to_squash) < 1:
            raise SquashError(
                "Invalid number of layers to squash: %s" % len(self.layers_to_squash))

        if len(self.layers_to_squash) == 1:
            raise SquashUnnecessaryError(
                "Single layer marked to squash, no squashing is required")

        self.log.info("Attempting to squash last %s layers...",
                      number_of_layers)
        self.log.debug("Layers to squash: %s", self.layers_to_squash)
        self.log.debug("Layers to move: %s", self.layers_to_move)

        # Fetch the image and unpack it on the fly to the old image directory
        self._save_image(self.old_image_id, self.old_image_dir)

        self.size_before = self._dir_size(self.old_image_dir)

        self.log.info("Squashing image '%s'..." % self.image)

    def _after_squashing(self):
        self.log.debug("Removing from disk already squashed layers...")
        shutil.rmtree(self.old_image_dir, ignore_errors=True)

        self.size_after = self._dir_size(self.new_image_dir)

        size_before_mb = float(self.size_before)/1024/1024
        size_after_mb = float(self.size_after)/1024/1024

        self.log.info("Original image size: %.2f MB" % size_before_mb)
        self.log.info("Squashed image size: %.2f MB" % size_after_mb)

        if (size_after_mb >= size_before_mb):
            self.log.info("If the squashed image is larger than original it means that there were no meaningful files to squash and it just added metadata. Are you sure you specified correct parameters?")
        else:
            self.log.info("Image size decreased by %.2f %%" % float(
                ((size_before_mb-size_after_mb)/size_before_mb)*100))

    def _dir_size(self, directory):
        size = 0

        for path, dirs, files in os.walk(directory):
            for f in files:
                size += os.path.getsize(os.path.join(path, f))

        return size

    def layer_paths(self):
        """
        Returns name of directories to layers in the exported tar archive.
        """
        pass

    def export_tar_archive(self, target_tar_file):
        self._tar_image(target_tar_file, self.new_image_dir)
        self.log.info("Image available at '%s'" % target_tar_file)

    def load_squashed_image(self):
        self._load_image(self.new_image_dir)

        if self.tag:
            self.log.info("Image registered in Docker daemon as %s:%s" %
                          (self.image_name, self.image_tag))

    def _files_in_layers(self, layers, directory):
        """
        Prepare a list of files in all layers
        """
        files = {}

        for layer in layers:
            self.log.debug("Generating list of files in layer '%s'..." % layer)
            tar_file = os.path.join(directory, layer, "layer.tar")
            with tarfile.open(tar_file, 'r', format=tarfile.PAX_FORMAT) as tar:
                files[layer] = [self._normalize_path(
                    x) for x in tar.getnames()]
            self.log.debug("Done, found %s files" % len(files[layer]))

        return files

    def _prepare_tmp_directory(self, tmp_dir):
        """ Creates temporary directory that is used to work on layers """

        if tmp_dir:
            if os.path.exists(tmp_dir):
                raise SquashError(
                    "The '%s' directory already exists, please remove it before you proceed" % tmp_dir)
            os.makedirs(tmp_dir)
        else:
            tmp_dir = tempfile.mkdtemp(prefix="docker-squash-")

        self.log.debug("Using %s as the temporary directory" % tmp_dir)

        return tmp_dir

    def _load_image(self, directory):

        tar_file = os.path.join(self.tmp_dir, "image.tar")

        self._tar_image(tar_file, directory)

        with open(tar_file, 'rb') as f:
            self.log.debug("Loading squashed image...")
            self.docker.load_image(f)
            self.log.debug("Image loaded!")

        os.remove(tar_file)

    def _tar_image(self, target_tar_file, directory):
        with tarfile.open(target_tar_file, 'w', format=tarfile.PAX_FORMAT) as tar:
            self.log.debug("Generating tar archive for the squashed image...")
            with Chdir(directory):
                # docker produces images like this:
                #   repositories
                #   <layer>/json
                # and not:
                #   ./
                #   ./repositories
                #   ./<layer>/json
                for f in os.listdir("."):
                    tar.add(f)
            self.log.debug("Archive generated")

    def _layers_to_squash(self, layers, from_layer):
        """ Prepares a list of layer IDs that should be squashed """
        to_squash = []
        to_leave = []
        should_squash = True

        for l in reversed(layers):
            if l == from_layer:
                should_squash = False

            if should_squash:
                to_squash.append(l)
            else:
                to_leave.append(l)

        to_squash.reverse()
        to_leave.reverse()

        return to_squash, to_leave

    def _extract_tar(self, fileobj, directory):
        with tarfile.open(fileobj=fileobj, mode='r|') as tar:
            tar.extractall(path=directory)

    def _save_image(self, image_id, directory):
        """ Saves the image as a tar archive under specified name """

        for x in [0, 1, 2]:
            self.log.info("Saving image %s to %s directory..." %
                          (image_id, directory))
            self.log.debug("Try #%s..." % (x + 1))

            try:
                image = self.docker.get_image(image_id)

                if docker.version_info[0] < 3:
                    # Docker library prior to 3.0.0 returned the requests
                    # object directly which cold be used to read from
                    self.log.debug(
                        "Extracting image using HTTPResponse object directly")
                    self._extract_tar(image, directory)
                else:
                    # Docker library >=3.0.0 returns iterator over raw data
                    self.log.debug(
                        "Extracting image using iterator over raw data")

                    fd_r, fd_w = os.pipe()

                    r = os.fdopen(fd_r, 'rb')
                    w = os.fdopen(fd_w, 'wb')

                    extracter = threading.Thread(
                        target=self._extract_tar, args=(r, directory))
                    extracter.start()

                    for chunk in image:
                        w.write(chunk)

                    w.flush()
                    w.close()

                    extracter.join()
                    r.close()
                self.log.info("Image saved!")
                return True
            except Exception as e:
                self.log.exception(e)
                self.log.warning(
                    "An error occured while saving the %s image, retrying..." % image_id)

        raise SquashError("Couldn't save %s image!" % image_id)

    def _unpack(self, tar_file, directory):
        """ Unpacks tar archive to selected directory """

        self.log.info("Unpacking %s tar file to %s directory" %
                      (tar_file, directory))

        with tarfile.open(tar_file, 'r') as tar:
            tar.extractall(path=directory)

        self.log.info("Archive unpacked!")

    def _read_layers(self, layers, image_id):
        """ Reads the JSON metadata for specified layer / image id """

        for layer in self.docker.history(image_id):
            layers.append(layer['Id'])

    def _parse_image_name(self, image):
        """
        Parses the provided image name and splits it in the
        name and tag part, if possible. If no tag is provided
        'latest' is used.
        """
        if ':' in image and '/' not in image.split(':')[-1]:
            image_tag = image.split(':')[-1]
            image_name = image[:-(len(image_tag) + 1)]
        else:
            image_tag = "latest"
            image_name = image

        return (image_name, image_tag)

    def _dump_json(self, data, new_line=False):
        """
        Helper function to marshal object into JSON string.
        Additionally a sha256sum of the created JSON string is generated.
        """

        # We do not want any spaces between keys and values in JSON
        json_data = json.dumps(data, separators=(',', ':'))

        if new_line:
            json_data = "%s\n" % json_data

        # Generate sha256sum of the JSON data, may be handy
        sha = hashlib.sha256(json_data.encode('utf-8')).hexdigest()

        return json_data, sha

    def _generate_repositories_json(self, repositories_file, image_id, name, tag):
        if not image_id:
            raise SquashError("Provided image id cannot be null")

        if name == tag == None:
            self.log.debug(
                "No name and tag provided for the image, skipping generating repositories file")
            return

        repos = {}
        repos[name] = {}
        repos[name][tag] = image_id

        data = json.dumps(repos, separators=(',', ':'))

        with open(repositories_file, 'w') as f:
            f.write(data)
            f.write("\n")

    def _write_version_file(self, squashed_dir):
        version_file = os.path.join(squashed_dir, "VERSION")

        with open(version_file, 'w') as f:
            f.write("1.0")

    def _write_json_metadata(self, metadata, metadata_file):
        with open(metadata_file, 'w') as f:
            f.write(metadata)

    def _read_old_metadata(self, old_json_file):
        self.log.debug("Reading JSON metadata file '%s'..." % old_json_file)

        # Read original metadata
        with open(old_json_file, 'r') as f:
            metadata = json.load(f)

        return metadata

    def _move_layers(self, layers, src, dest):
        """
        This moves all the layers that should be copied as-is.
        In other words - all layers that are not meant to be squashed will be
        moved from the old image to the new image untouched.
        """
        for layer in layers:
            layer_id = layer.replace('sha256:', '')

            self.log.debug("Moving unmodified layer '%s'..." % layer_id)
            shutil.move(os.path.join(src, layer_id), dest)

    def _file_should_be_skipped(self, file_name, file_paths):
        # file_paths is now array of array with files to be skipped.
        # First level are layers, second are files in these layers.
        layer_nb = 1

        for layers in file_paths:
            for file_path in layers:
                if file_name == file_path or file_name.startswith(file_path + "/"):
                    return layer_nb

            layer_nb += 1

        return 0

    def _marker_files(self, tar, members):
        """
        Searches for marker files in the specified archive.

        Docker marker files are files taht have the .wh. prefix in the name.
        These files mark the corresponding file to be removed (hidden) when
        we start a container from the image.
        """
        marker_files = {}

        self.log.debug(
            "Searching for marker files in '%s' archive..." % tar.name)

        for member in members:
            if '.wh.' in member.name:
                self.log.debug("Found '%s' marker file" % member.name)
                marker_files[member] = tar.extractfile(member)

        self.log.debug("Done, found %s files" % len(marker_files))

        return marker_files

    def _add_markers(self, markers, tar, files_in_layers, added_symlinks):
        """
        This method is responsible for adding back all markers that were not
        added to the squashed layer AND files they refer to can be found in layers
        we do not squash.
        """

        if markers:
            self.log.debug("Marker files to add: %s" %
                           [o.name for o in markers.keys()])
        else:
            # No marker files to add
            return

        # https://github.com/goldmann/docker-squash/issues/108
        # Some tar archives do have the filenames prefixed with './'
        # which does not have any effect when we unpack the tar achive,
        # but when processing tar content - we see this.
        tar_files = [self._normalize_path(x) for x in tar.getnames()]

        for marker, marker_file in six.iteritems(markers):
            actual_file = marker.name.replace('.wh.', '')
            normalized_file = self._normalize_path(actual_file)

            should_be_added_back = False

            if self._file_should_be_skipped(normalized_file, added_symlinks):
                self.log.debug(
                    "Skipping '%s' marker file, this file is on a symlink path" % normalized_file)
                continue

            if normalized_file in tar_files:
                self.log.debug(
                    "Skipping '%s' marker file, this file was added earlier for some reason..." % normalized_file)
                continue

            if files_in_layers:
                for files in files_in_layers.values():
                    if normalized_file in files:
                        should_be_added_back = True
                        break
            else:
                # There are no previous layers, so we need to add it back
                # In fact this shouldn't happen since having a marker file
                # where there is no previous layer does not make sense.
                should_be_added_back = True

            if should_be_added_back:
                self.log.debug(
                    "Adding '%s' marker file back..." % marker.name)
                # Marker files on AUFS are hardlinks, we need to create
                # regular files, therefore we need to recreate the tarinfo
                # object
                tar.addfile(tarfile.TarInfo(name=marker.name), marker_file)
                # Add the file name to the list too to avoid re-reading all files
                # in tar archive
                tar_files.append(normalized_file)
            else:
                self.log.debug(
                    "Skipping '%s' marker file..." % marker.name)

    def _normalize_path(self, path):
        return os.path.normpath(os.path.join("/", path))

    def _add_hardlinks(self, squashed_tar, squashed_files, to_skip, skipped_hard_links):
        for layer, hardlinks_in_layer in enumerate(skipped_hard_links):
            # We need to start from 1, that's why we bump it here
            current_layer = layer + 1
            for member in six.itervalues(hardlinks_in_layer):
                normalized_name = self._normalize_path(member.name)
                normalized_linkname = self._normalize_path(member.linkname)

                # Find out if the name is on the list of files to skip - if it is - get the layer number
                # where it was found
                layer_skip_name = self._file_should_be_skipped(
                    normalized_name, to_skip)
                # Do the same for linkname
                layer_skip_linkname = self._file_should_be_skipped(
                    normalized_linkname, to_skip)

                # We need to check if we should skip adding back the hard link
                # This can happen in the following situations:
                # 1. hard link is on the list of files to skip
                # 2. hard link target is on the list of files to skip
                # 3. hard link is already in squashed files
                # 4. hard link target is NOT in already squashed files
                if layer_skip_name and current_layer > layer_skip_name or layer_skip_linkname and current_layer > layer_skip_linkname or normalized_name in squashed_files or normalized_linkname not in squashed_files:
                    self.log.debug("Found a hard link '%s' to a file which is marked to be skipped: '%s', skipping link too" % (
                        normalized_name, normalized_linkname))
                else:
                    if self.debug:
                        self.log.debug("Adding hard link '%s' pointing to '%s' back..." % (
                            normalized_name, normalized_linkname))

                    squashed_files.append(normalized_name)
                    squashed_tar.addfile(member)

    def _add_file(self, member, content, squashed_tar, squashed_files, to_skip):
        normalized_name = self._normalize_path(member.name)

        if normalized_name in squashed_files:
            self.log.debug(
                "Skipping file '%s' because it is already squashed" % normalized_name)
            return

        if self._file_should_be_skipped(normalized_name, to_skip):
            self.log.debug(
                "Skipping '%s' file because it's on the list to skip files" % normalized_name)
            return

        if content:
            squashed_tar.addfile(member, content)
        else:
            # Special case: other(?) files, we skip the file
            # itself
            squashed_tar.addfile(member)

        # We added a file to the squashed tar, so let's note it
        squashed_files.append(normalized_name)

    def _add_symlinks(self, squashed_tar, squashed_files, to_skip, skipped_sym_links):
        added_symlinks = []
        for layer, symlinks_in_layer in enumerate(skipped_sym_links):
            # We need to start from 1, that's why we bump it here
            current_layer = layer + 1
            for member in six.itervalues(symlinks_in_layer):

                # Handling symlinks. This is similar to hard links with one
                # difference. Sometimes we do want to have broken symlinks
                # be addedeither case because these can point to locations
                # that will become avaialble after adding volumes for example.
                normalized_name = self._normalize_path(member.name)
                normalized_linkname = self._normalize_path(member.linkname)

                # File is already in squashed files, skipping
                if normalized_name in squashed_files:
                    self.log.debug(
                        "Found a symbolic link '%s' which is already squashed, skipping" % (normalized_name))
                    continue

                if self._file_should_be_skipped(normalized_name, added_symlinks):
                    self.log.debug(
                        "Found a symbolic link '%s' which is on a path to previously squashed symlink, skipping" % (normalized_name))
                    continue
                # Find out if the name is on the list of files to skip - if it is - get the layer number
                # where it was found
                layer_skip_name = self._file_should_be_skipped(
                    normalized_name, to_skip)
                # Do the same for linkname
                layer_skip_linkname = self._file_should_be_skipped(
                    normalized_linkname, to_skip)

                # If name or linkname was found in the lists of files to be
                # skipped or it's not found in the squashed files
                if layer_skip_name and current_layer > layer_skip_name or layer_skip_linkname and current_layer > layer_skip_linkname:
                    self.log.debug("Found a symbolic link '%s' to a file which is marked to be skipped: '%s', skipping link too" % (
                        normalized_name, normalized_linkname))
                else:
                    if self.debug:
                        self.log.debug("Adding symbolic link '%s' pointing to '%s' back..." % (
                            normalized_name, normalized_linkname))

                    added_symlinks.append([normalized_name])

                    squashed_files.append(normalized_name)
                    squashed_tar.addfile(member)

        return added_symlinks

    def _squash_layers(self, layers_to_squash, layers_to_move):
        self.log.info("Starting squashing...")

        # Reverse the layers to squash - we begin with the newest one
        # to make the tar lighter
        layers_to_squash.reverse()

        # Find all files in layers that we don't squash
        files_in_layers_to_move = self._files_in_layers(
            layers_to_move, self.old_image_dir)

        with tarfile.open(self.squashed_tar, 'w', format=tarfile.PAX_FORMAT) as squashed_tar:
            to_skip = []
            skipped_markers = {}
            skipped_hard_links = []
            skipped_sym_links = []
            skipped_files = []
            # List of filenames in the squashed archive
            squashed_files = []
            # List of opaque directories in the image
            opaque_dirs = []

            for layer_id in layers_to_squash:
                layer_tar_file = os.path.join(
                    self.old_image_dir, layer_id, "layer.tar")

                self.log.info("Squashing file '%s'..." % layer_tar_file)

                # Open the exiting layer to squash
                with tarfile.open(layer_tar_file, 'r', format=tarfile.PAX_FORMAT) as layer_tar:
                    # Find all marker files for all layers
                    # We need the list of marker files upfront, so we can
                    # skip unnecessary files
                    members = layer_tar.getmembers()
                    markers = self._marker_files(layer_tar, members)

                    skipped_sym_link_files = {}
                    skipped_hard_link_files = {}
                    skipped_files_in_layer = {}

                    files_to_skip = []
                    # List of opaque directories found in this layer
                    layer_opaque_dirs = []

                    # Add it as early as possible, we will be populating
                    # 'skipped_sym_link_files' array later
                    skipped_sym_links.append(skipped_sym_link_files)

                    # Add it as early as possible, we will be populating
                    # 'files_to_skip' array later
                    to_skip.append(files_to_skip)

                    # Iterate over marker files found for this particular
                    # layer and if a file in the squashed layers file corresponding
                    # to the marker file is found, then skip both files
                    for marker, marker_file in six.iteritems(markers):
                        # We have a opaque directory marker file
                        # https://github.com/opencontainers/image-spec/blob/master/layer.md#opaque-whiteout
                        if marker.name.endswith('.wh..wh..opq'):
                            opaque_dir = os.path.dirname(marker.name)

                            self.log.debug(
                                "Found opaque directory: '%s'" % opaque_dir)

                            layer_opaque_dirs.append(opaque_dir)
                        else:
                            files_to_skip.append(
                                self._normalize_path(marker.name.replace('.wh.', '')))

                            skipped_markers[marker] = marker_file

                    # Copy all the files to the new tar
                    for member in members:
                        normalized_name = self._normalize_path(member.name)

                        if self._is_in_opaque_dir(member, opaque_dirs):
                            self.log.debug(
                                "Skipping file '%s' because it is in an opaque directory" % normalized_name)
                            continue

                        # Skip all symlinks, we'll investigate them later
                        if member.issym():
                            skipped_sym_link_files[normalized_name] = member
                            continue

                        if member in six.iterkeys(skipped_markers):
                            self.log.debug(
                                "Skipping '%s' marker file, at the end of squashing we'll see if it's necessary to add it back" % normalized_name)
                            continue

                        if self._file_should_be_skipped(normalized_name, skipped_sym_links):
                            self.log.debug(
                                "Skipping '%s' file because it's on a symlink path, at the end of squashing we'll see if it's necessary to add it back" % normalized_name)

                            if member.isfile():
                                f = (member, layer_tar.extractfile(member))
                            else:
                                f = (member, None)

                            skipped_files_in_layer[normalized_name] = f
                            continue

                        # Skip files that are marked to be skipped
                        if self._file_should_be_skipped(normalized_name, to_skip):
                            self.log.debug(
                                "Skipping '%s' file because it's on the list to skip files" % normalized_name)
                            continue

                        # Check if file is already added to the archive
                        if normalized_name in squashed_files:
                            # File already exist in the squashed archive, skip it because
                            # file want to add is older than the one already in the archive.
                            # This is true because we do reverse squashing - from
                            # newer to older layer
                            self.log.debug(
                                "Skipping '%s' file because it's older than file already added to the archive" % normalized_name)
                            continue

                        # Hard links are processed after everything else
                        if member.islnk():
                            skipped_hard_link_files[normalized_name] = member
                            continue

                        content = None

                        if member.isfile():
                            content = layer_tar.extractfile(member)

                        self._add_file(member, content,
                                       squashed_tar, squashed_files, to_skip)

                    skipped_hard_links.append(skipped_hard_link_files)
                    skipped_files.append(skipped_files_in_layer)
                    opaque_dirs += layer_opaque_dirs

            self._add_hardlinks(squashed_tar, squashed_files,
                                to_skip, skipped_hard_links)
            added_symlinks = self._add_symlinks(
                squashed_tar, squashed_files, to_skip, skipped_sym_links)

            for layer in skipped_files:
                for member, content in six.itervalues(layer):
                    self._add_file(member, content, squashed_tar,
                                   squashed_files, added_symlinks)

            if files_in_layers_to_move:
                self._reduce(skipped_markers)

                self._add_markers(skipped_markers, squashed_tar,
                                  files_in_layers_to_move, added_symlinks)

        self.log.info("Squashing finished!")

    def _is_in_opaque_dir(self, member, dirs):
        """
        If the member we investigate is an opaque directory
        or if the member is located inside of the opaque directory,
        we copy these files as-is. Any other layer that has content
        on the opaque directory will be ignored!
        """

        for opaque_dir in dirs:
            if member.name == opaque_dir or member.name.startswith("%s/" % opaque_dir):
                self.log.debug("Member '%s' found to be part of opaque directory '%s'" % (
                    member.name, opaque_dir))
                return True

        return False

    def _reduce(self, markers):
        """
        This function is responsible for reducing marker files
        that are scheduled to be added at the end of squashing to
        minimum.

        In some cases, one marker file will overlap
        with others making others not necessary.

        This is not only about adding less marker files, but
        if we try to add a marker file for a file or directory
        deeper in the hierarchy of already marked directory,
        the image will not be successfully loaded back into Docker
        daemon.

        Passed dictionary containing markers is altered *in-place*.

        Args:
            markers (dict): Dictionary of markers scheduled to be added.
        """

        self.log.debug("Reducing marker files to be added back...")

        # Prepare a list of files (or directories) based on the marker
        # files scheduled to be added
        marked_files = list(map(lambda x: self._normalize_path(
            x.name.replace('.wh.', '')), markers.keys()))

        # List of markers that should be not added back to tar file
        to_remove = []

        for marker in markers.keys():
            self.log.debug(
                "Investigating '{}' marker file".format(marker.name))

            path = self._normalize_path(marker.name.replace('.wh.', ''))
            # Iterate over the path hierarchy, but starting with the
            # root directory. This will make it possible to remove
            # marker files based on the highest possible directory level
            for directory in reversed(self._path_hierarchy(path)):
                if directory in marked_files:
                    self.log.debug(
                        "Marker file '{}' is superseded by higher-level marker file: '{}'".format(marker.name, directory))
                    to_remove.append(marker)
                    break

        self.log.debug("Removing {} marker files".format(len(to_remove)))

        if to_remove:
            for marker in to_remove:
                self.log.debug("Removing '{}' marker file".format(marker.name))
                markers.pop(marker)

        self.log.debug("Marker files reduced")

    def _path_hierarchy(self, path):
        """
        Creates a full hierarchy of directories for a given path.

        For a particular path, a list will be returned
        containing paths from the path specified, through all levels
        up to the root directory.

        Example:
            Path '/opt/testing/some/dir/structure/file'

            will return:

            ['/opt/testing/some/dir/structure', '/opt/testing/some/dir', '/opt/testing/some', '/opt/testing', '/opt', '/']
        """
        if not path:
            raise SquashError("No path provided to create the hierarchy for")

        hierarchy = []

        dirname = os.path.dirname(path)

        hierarchy.append(dirname)

        # If we are already at root level, stop
        if dirname != '/':
            hierarchy.extend(self._path_hierarchy(dirname))

        return hierarchy



import hashlib
import os
import random
import shutil

from docker_squash.image import Image


class V1Image(Image):
    FORMAT = 'v1'

    def _generate_image_id(self):
        while True:
            image_id = hashlib.sha256(
                str(random.getrandbits(128)).encode('utf8')).hexdigest()

            try:
                int(image_id[0:10])
            except ValueError:
                # All good!
                return image_id

    def _before_squashing(self):
        super(V1Image, self)._before_squashing()

        if self.layers_to_move:
            self.squash_id = self.layers_to_move[-1]

    def _squash(self):
        # Prepare the directory
        os.makedirs(self.squashed_dir)
        self._squash_layers(self.layers_to_squash, self.layers_to_move)
        self._write_version_file(self.squashed_dir)
        # Move all the layers that should be untouched
        self._move_layers(self.layers_to_move,
                          self.old_image_dir, self.new_image_dir)

        config_file = os.path.join(
            self.old_image_dir, self.old_image_id, "json")

        image_id = self._update_squashed_layer_metadata(
            config_file, self.squashed_dir)
        shutil.move(self.squashed_dir, os.path.join(
            self.new_image_dir, image_id))
        repositories_file = os.path.join(self.new_image_dir, "repositories")
        self._generate_repositories_json(
            repositories_file, image_id, self.image_name, self.image_tag)

        return image_id

    def _update_squashed_layer_metadata(self, old_json_file, squashed_dir):
        image_id = self._generate_image_id()

        metadata = self._read_old_metadata(old_json_file)

        # Modify common metadata fields
        if self.squash_id:
            metadata['config']['Image'] = self.squash_id
        else:
            metadata['config'].pop('Image', None)

        if 'parent_id' in metadata and self.squash_id:
            metadata['parent_id'] = "sha256:%s" % self.squash_id
        else:
            metadata.pop('parent_id', None)

        metadata.pop('layer_id', None)

        metadata['created'] = self.date

        # Remove unnecessary fields
        del metadata['container_config']
        del metadata['container']
        del metadata['config']['Hostname']

        if self.squash_id:
            metadata['parent'] = self.squash_id
        else:
            metadata.pop('parent', None)

        metadata['id'] = image_id
        metadata['Size'] = os.path.getsize(
            os.path.join(squashed_dir, "layer.tar"))
        json_metadata = self._dump_json(metadata)[0]

        self._write_json_metadata(
            "%s" % json_metadata, os.path.join(squashed_dir, "json"))

        return image_id



# -*- coding: utf-8 -*-

import argparse
import logging
import sys

from docker_squash import squash
from docker_squash.errors import SquashError
from docker_squash.version import version


# Source: http://stackoverflow.com/questions/1383254/logging-streamhandler-and-standard-streams
class SingleLevelFilter(logging.Filter):
    def __init__(self, passlevel, reject):
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return (record.levelno != self.passlevel)
        else:
            return (record.levelno == self.passlevel)


class MyParser(argparse.ArgumentParser):

    def error(self, message):
        self.print_help()
        sys.stderr.write('\nError: %s\n' % message)
        sys.exit(2)


class CLI(object):

    def __init__(self):
        handler_out = logging.StreamHandler(sys.stdout)
        handler_err = logging.StreamHandler(sys.stderr)

        handler_out.addFilter(SingleLevelFilter(logging.INFO, False))
        handler_err.addFilter(SingleLevelFilter(logging.INFO, True))

        self.log = logging.getLogger()
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

        handler_out.setFormatter(formatter)
        handler_err.setFormatter(formatter)

        self.log.addHandler(handler_out)
        self.log.addHandler(handler_err)

    def run(self):
        parser = MyParser(
            description='Docker layer squashing tool')

        parser.add_argument(
            '-v', '--verbose', action='store_true', help='Verbose output')

        parser.add_argument(
            '--version', action='version', help='Show version and exit', version=version)

        parser.add_argument('image', help='Image to be squashed')
        parser.add_argument(
            '-d', '--development', action='store_true', help='Does not clean up after failure for easier debugging')
        parser.add_argument(
            '-f', '--from-layer', help='Number of layers to squash or ID of the layer (or image ID or image name) to squash from. In case the provided value is an integer, specified number of layers will be squashed. Every layer in the image will be squashed if the parameter is not provided.')
        parser.add_argument(
            '-t', '--tag', help="Specify the tag to be used for the new image. If not specified no tag will be applied")
        parser.add_argument(
            '-m', '--message', help="Specify a commit message (comment) for the new image.")
        parser.add_argument(
            '-c', '--cleanup', action='store_true', help="Remove source image from Docker after squashing")
        parser.add_argument(
            '--tmp-dir', help='Temporary directory to be created and used')
        parser.add_argument(
            '--output-path', help='Path where the image should be stored after squashing. If not provided, image will be loaded into Docker daemon')

        args = parser.parse_args()

        if args.verbose:
            self.log.setLevel(logging.DEBUG)
        else:
            self.log.setLevel(logging.INFO)

        self.log.debug("Running version %s", version)

        try:
            squash.Squash(log=self.log, image=args.image,
                          from_layer=args.from_layer, tag=args.tag, comment=args.message, output_path=args.output_path, tmp_dir=args.tmp_dir, development=args.development, cleanup=args.cleanup).run()
        except KeyboardInterrupt:
            self.log.error("Program interrupted by user, exiting...")
            sys.exit(1)
        except:
            e = sys.exc_info()[1]

            if args.development or args.verbose:
                self.log.exception(e)
            else:
                self.log.error(str(e))

            self.log.error(
                "Execution failed, consult logs above. If you think this is our fault, please file an issue: https://github.com/goldmann/docker-squash/issues, thanks!")

            if isinstance(e, SquashError):
                sys.exit(e.code)

            sys.exit(1)


def run():
    cli = CLI()
    cli.run()


if __name__ == "__main__":
    run()




import hashlib
import json
import os
import shutil

from collections import OrderedDict
from docker_squash.image import Image


class V2Image(Image):
    FORMAT = 'v2'

    def _before_squashing(self):
        super(V2Image, self)._before_squashing()

        # Read old image manifest file
        self.old_image_manifest = self._read_json_file(
            os.path.join(self.old_image_dir, "manifest.json"))[0]

        # Read old image config file
        self.old_image_config = self._read_json_file(os.path.join(
            self.old_image_dir, self.old_image_manifest['Config']))

        # Read layer paths inside of the tar archive
        # We split it into layers that needs to be squashed
        # and layers that needs to be moved as-is
        self.layer_paths_to_squash, self.layer_paths_to_move = self._read_layer_paths(
            self.old_image_config, self.old_image_manifest, self.layers_to_move)

        if self.layer_paths_to_move:
            self.squash_id = self.layer_paths_to_move[-1]

    def _squash(self):
        if self.layer_paths_to_squash:
            # Prepare the directory
            os.makedirs(self.squashed_dir)
            # Merge data layers
            self._squash_layers(self.layer_paths_to_squash,
                                self.layer_paths_to_move)

        self.diff_ids = self._generate_diff_ids()
        self.chain_ids = self._generate_chain_ids(self.diff_ids)

        metadata = self._generate_image_metadata()
        image_id = self._write_image_metadata(metadata)

        layer_path_id = None

        if self.layer_paths_to_squash:
            # Compute layer id to use to name the directory where
            # we store the layer data inside of the tar archive
            layer_path_id = self._generate_squashed_layer_path_id()

            metadata = self._generate_last_layer_metadata(
                layer_path_id, self.layer_paths_to_squash[0])
            self._write_squashed_layer_metadata(metadata)

            # Write version file to the squashed layer
            # Even Docker doesn't know why it's needed...
            self._write_version_file(self.squashed_dir)

            # Move the temporary squashed layer directory to the correct one
            shutil.move(self.squashed_dir, os.path.join(
                self.new_image_dir, layer_path_id))

        manifest = self._generate_manifest_metadata(
            image_id, self.image_name, self.image_tag, self.old_image_manifest, self.layer_paths_to_move, layer_path_id)

        self._write_manifest_metadata(manifest)

        repository_image_id = manifest[0]["Layers"][-1].split("/")[0]

        # Move all the layers that should be untouched
        self._move_layers(self.layer_paths_to_move,
                          self.old_image_dir, self.new_image_dir)

        repositories_file = os.path.join(self.new_image_dir, "repositories")
        self._generate_repositories_json(
            repositories_file, repository_image_id, self.image_name, self.image_tag)

        return image_id

    def _write_image_metadata(self, metadata):
        # Create JSON from the metadata
        # Docker adds new line at the end
        json_metadata, image_id = self._dump_json(metadata, True)
        image_metadata_file = os.path.join(
            self.new_image_dir, "%s.json" % image_id)

        self._write_json_metadata(json_metadata, image_metadata_file)

        return image_id

    def _write_squashed_layer_metadata(self, metadata):
        layer_metadata_file = os.path.join(self.squashed_dir, "json")
        json_metadata = self._dump_json(metadata)[0]

        self._write_json_metadata(json_metadata, layer_metadata_file)

    def _write_manifest_metadata(self, manifest):
        manifest_file = os.path.join(self.new_image_dir, "manifest.json")
        json_manifest = self._dump_json(manifest, True)[0]

        self._write_json_metadata(json_manifest, manifest_file)

    def _generate_manifest_metadata(self, image_id, image_name, image_tag, old_image_manifest, layer_paths_to_move, layer_path_id=None):
        manifest = OrderedDict()
        manifest['Config'] = "%s.json" % image_id

        if image_name and image_tag:
            manifest['RepoTags'] = ["%s:%s" % (image_name, image_tag)]

        manifest['Layers'] = old_image_manifest[
            'Layers'][:len(layer_paths_to_move)]

        if layer_path_id:
            manifest['Layers'].append("%s/layer.tar" % layer_path_id)

        return [manifest]

    def _read_json_file(self, json_file):
        """ Helper function to read JSON file as OrderedDict """

        self.log.debug("Reading '%s' JSON file..." % json_file)

        with open(json_file, 'r') as f:
            return json.load(f, object_pairs_hook=OrderedDict)

    def _read_layer_paths(self, old_image_config, old_image_manifest, layers_to_move):
        """
        In case of v2 format, layer id's are not the same as the id's
        used in the exported tar archive to name directories for layers.
        These id's can be found in the configuration files saved with
        the image - we need to read them.
        """

        # In manifest.json we do not have listed all layers
        # but only layers that do contain some data.
        current_manifest_layer = 0

        layer_paths_to_move = []
        layer_paths_to_squash = []

        # Iterate over image history, from base image to top layer
        for i, layer in enumerate(old_image_config['history']):

            # If it's not an empty layer get the id
            # (directory name) where the layer's data is
            # stored
            if not layer.get('empty_layer', False):
                layer_id = old_image_manifest['Layers'][
                    current_manifest_layer].rsplit('/')[0]

                # Check if this layer should be moved or squashed
                if len(layers_to_move) > i:
                    layer_paths_to_move.append(layer_id)
                else:
                    layer_paths_to_squash.append(layer_id)

                current_manifest_layer += 1

        return layer_paths_to_squash, layer_paths_to_move

    def _generate_chain_id(self, chain_ids, diff_ids, parent_chain_id):
        if parent_chain_id == None:
            return self._generate_chain_id(chain_ids, diff_ids[1:], diff_ids[0])

        chain_ids.append(parent_chain_id)

        if len(diff_ids) == 0:
            return parent_chain_id

        # This probably should not be hardcoded
        to_hash = "sha256:%s sha256:%s" % (parent_chain_id, diff_ids[0])
        digest = hashlib.sha256(str(to_hash).encode('utf8')).hexdigest()

        return self._generate_chain_id(chain_ids, diff_ids[1:], digest)

    def _generate_chain_ids(self, diff_ids):
        chain_ids = []

        self._generate_chain_id(chain_ids, diff_ids, None)

        return chain_ids

    def _generate_diff_ids(self):
        diff_ids = []

        for path in self.layer_paths_to_move:
            sha256 = self._compute_sha256(os.path.join(self.old_image_dir, path, "layer.tar"))
            diff_ids.append(sha256)

        if self.layer_paths_to_squash:
            sha256 = self._compute_sha256(os.path.join(self.squashed_dir, "layer.tar"))
            diff_ids.append(sha256)

        return diff_ids

    def _compute_sha256(self, layer_tar):
        sha256 = hashlib.sha256()

        with open(layer_tar, 'rb') as f:
            while True:
                # Read in 10MB chunks
                data = f.read(10485760)

                if not data:
                    break

                sha256.update(data)

        return sha256.hexdigest()

    def _generate_squashed_layer_path_id(self):
        """
        This function generates the id used to name the directory to
        store the squashed layer content in the archive.

        This mimics what Docker does here: https://github.com/docker/docker/blob/v1.10.0-rc1/image/v1/imagev1.go#L42
        To make it simpler we do reuse old image metadata and
        modify it to what it should look which means to be exact
        as https://github.com/docker/docker/blob/v1.10.0-rc1/image/v1/imagev1.go#L64
        """

        # Using OrderedDict, because order of JSON elements is important
        v1_metadata = OrderedDict(self.old_image_config)

        # Update image creation date
        v1_metadata['created'] = self.date

        # Remove unnecessary elements
        # Do not fail if key is not found
        for key in 'history', 'rootfs', 'container':
            v1_metadata.pop(key, None)

        # Docker internally changes the order of keys between
        # exported metadata (why oh why?!). We need to add 'os'
        # element after 'layer_id'
        operating_system = v1_metadata.pop('os', None)

        # The 'layer_id' element is the chain_id of the
        # squashed layer
        v1_metadata['layer_id'] = "sha256:%s" % self.chain_ids[-1]

        # Add back 'os' element
        if operating_system:
            v1_metadata['os'] = operating_system

        # The 'parent' element is the name of the directory (inside the
        # exported tar archive) of the last layer that we move
        # (layer below squashed layer)

        if self.layer_paths_to_move:
            if self.layer_paths_to_squash:
                parent = self.layer_paths_to_move[-1]
            else:
                parent = self.layer_paths_to_move[0]

            v1_metadata['parent'] = "sha256:%s" % parent

        # The 'Image' element is the id of the layer from which we squash
        if self.squash_id:
            # Update image id, should be one layer below squashed layer
            v1_metadata['config']['Image'] = self.squash_id
        else:
            v1_metadata['config']['Image'] = ""

        # Get the sha256sum of the JSON exported metadata,
        # we do not care about the metadata anymore
        sha = self._dump_json(v1_metadata)[1]

        return sha

    def _generate_last_layer_metadata(self, layer_path_id, old_layer_path=None):
        if not old_layer_path:
            old_layer_path = layer_path_id

        config_file = os.path.join(
            self.old_image_dir, old_layer_path, "json")

        with open(config_file, 'r') as f:
            config = json.load(f, object_pairs_hook=OrderedDict)

        config['created'] = self.date

        if self.squash_id:
            # Update image id, should be one layer below squashed layer
            config['config']['Image'] = self.squash_id
        else:
            config['config']['Image'] = ""

        # Update 'parent' - it should be path to the last layer to move
        if self.layer_paths_to_move:
            config['parent'] = self.layer_paths_to_move[-1]
        else:
            config.pop("parent", None)
        # Update 'id' - it should be the path to the layer
        config['id'] = layer_path_id
        config.pop("container", None)
        return config

    def _generate_image_metadata(self):
        # First - read old image config, we'll update it instead of
        # generating one from scratch
        metadata = OrderedDict(self.old_image_config)
        # Update image creation date
        metadata['created'] = self.date

        # Remove unnecessary or old fields
        metadata.pop("container", None)

        # Remove squashed layers from history
        metadata['history'] = metadata['history'][:len(self.layers_to_move)]
        # Remove diff_ids for squashed layers
        metadata['rootfs']['diff_ids'] = metadata['rootfs'][
            'diff_ids'][:len(self.layer_paths_to_move)]

        history = {'comment': self.comment, 'created': self.date}

        if self.layer_paths_to_squash:
            # Add diff_ids for the squashed layer
            metadata['rootfs']['diff_ids'].append(
                "sha256:%s" % self.diff_ids[-1])
        else:
            history['empty_layer'] = True

        # Add new entry for squashed layer to history
        metadata['history'].append(history)

        if self.squash_id:
            # Update image id, should be one layer below squashed layer
            metadata['config']['Image'] = self.squash_id
        else:
            metadata['config']['Image'] = ""

        return metadata



# -*- coding: utf-8 -*-

import os

import docker

from distutils.version import StrictVersion

from docker_squash.v1_image import V1Image
from docker_squash.v2_image import V2Image
from docker_squash.lib import common
from docker_squash.errors import SquashError
from docker_squash.version import version


class Squash(object):

    def __init__(self, log, image, docker=None, from_layer=None, tag=None, comment="", tmp_dir=None,
                 output_path=None, load_image=True, development=False, cleanup=False):
        self.log = log
        self.docker = docker
        self.image = image
        self.from_layer = from_layer
        self.tag = tag
        self.comment = comment
        self.tmp_dir = tmp_dir
        self.output_path = output_path
        self.load_image = load_image
        self.development = development
        self.cleanup = cleanup

        if not docker:
            self.docker = common.docker_client(self.log)

    def run(self):
        docker_version = self.docker.version()
        self.log.info("docker-squash version %s, Docker %s, API %s..." %
                      (version, docker_version['GitCommit'], docker_version['ApiVersion']))

        if self.image is None:
            raise SquashError("Image is not provided")

        if not (self.output_path or self.load_image):
            self.log.warning(
                "No output path specified and loading into Docker is not selected either; squashed image would not accessible, proceeding with squashing doesn't make sense")
            return

        if self.output_path and os.path.exists(self.output_path):
            self.log.warning(
                "Path '%s' specified as output path where the squashed image should be saved already exists, it'll be overriden" % self.output_path)

        if StrictVersion(docker_version['ApiVersion']) >= StrictVersion("1.22"):
            image = V2Image(self.log, self.docker, self.image,
                            self.from_layer, self.tmp_dir, self.tag, self.comment)
        else:
            image = V1Image(self.log, self.docker, self.image,
                            self.from_layer, self.tmp_dir, self.tag)

        self.log.info("Using %s image format" % image.FORMAT)

        try:
            return self.squash(image)
        except:
            # https://github.com/goldmann/docker-scripts/issues/44
            # If development mode is not enabled, make sure we clean up the
            # temporary directory
            if not self.development:
                image.cleanup()

            raise

    def _cleanup(self):
        try:
            image_id = self.docker.inspect_image(self.image)['Id']
        except docker.errors.APIError as ex:
            self.log.warning(
                "Could not get the image ID for {} image: {}, skipping cleanup after squashing".format(
                    self.image, str(ex)))
            return

        self.log.info("Removing old {} image...".format(self.image))

        try:
            self.docker.remove_image(image_id, force=False, noprune=False)
            self.log.info("Image {} removed!".format(self.image))
        except docker.errors.APIError as ex:
            self.log.warning(
                "Could not remove image {}: {}, skipping cleanup after squashing".format(self.image, str(ex)))

    def squash(self, image):
        # Do the actual squashing
        new_image_id=image.squash()

        self.log.info("New squashed image ID is %s" % new_image_id)

        if self.output_path:
            # Move the tar archive to the specified path
            image.export_tar_archive(self.output_path)

        if self.load_image:
            # Load squashed image into Docker
            image.load_squashed_image()

        # Clean up all temporary files
        image.cleanup()

        # Remove the source image - this is the only possible way
        # to remove orphaned layers from Docker daemon at the build time.
        # We cannot use here a tag name because it could be used as the target,
        # squashed image tag - we need to use the image ID.
        if self.cleanup:
            self._cleanup()

        self.log.info("Done")

        return new_image_id



# -*- coding: utf-8 -*-

import docker
import os
import requests

from docker_squash.errors import Error

# First try to import Docker client using the API
# available in version 2 of the library and fall
# back to version 1
try:
    from docker.api.client import APIClient as APIClientClass
except ImportError:
    from docker.client import Client as APIClientClass

DEFAULT_TIMEOUT_SECONDS = 600


def docker_client(log):
    log.debug("Preparing Docker client...")

    # Default timeout 10 minutes
    try:
        timeout = int(os.getenv('DOCKER_TIMEOUT', 600))
    except ValueError as e:
        raise Error("Provided timeout value: %s cannot be parsed as integer, exiting." %
                    os.getenv('DOCKER_TIMEOUT'))

    if not timeout > 0:
        raise Error(
            "Provided timeout value needs to be greater than zero, currently: %s, exiting." % timeout)

    # backwards compat
    try:
        os.environ["DOCKER_HOST"] = os.environ["DOCKER_CONNECTION"]
        log.warn("DOCKER_CONNECTION is deprecated, please use DOCKER_HOST instead")
    except KeyError:
        pass

    params = {"version": "auto"}
    params.update(docker.utils.kwargs_from_env())
    params["timeout"] = timeout

    try:
        client = APIClientClass(**params)
    except docker.errors.DockerException as e:
        log.error(
            "Could not create Docker client, please make sure that you specified valid parameters in the 'DOCKER_HOST' environment variable.")
        raise Error("Error while creating the Docker client: %s" % e)

    if client and valid_docker_connection(client):
        log.debug("Docker client ready")
        return client
    else:
        log.error(
            "Could not connect to the Docker daemon, please make sure the Docker daemon is running.")

        if os.environ.get('DOCKER_HOST'):
            log.error(
                "If Docker daemon is running, please make sure that you specified valid parameters in the 'DOCKER_HOST' environment variable.")

        raise Error("Cannot connect to Docker daemon")


def valid_docker_connection(client):
    try:
        return client.ping()
    except requests.exceptions.ConnectionError:
        return False



"""
This is a monkey patching for Python 2 that is required to handle PAX headers
in TAR files that are not decodable to UTF8. It leaves it undecoded and when
adding back to the tar archive the header is not encoded preserving the
original headers.

Reported in RH Bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1194473

Original source code was taken from Python 2.7.9.

Credit goes to Vincent Batts:

  https://github.com/docker/docker-registry/pull/381
"""

import re
import tarfile


def _proc_pax(self, filetar):
    """Process an extended or global header as described in POSIX.1-2001."""
    # Read the header information.
    buf = filetar.fileobj.read(self._block(self.size))

    # A pax header stores supplemental information for either
    # the following file (extended) or all following files
    # (global).
    if self.type == tarfile.XGLTYPE:
        pax_headers = filetar.pax_headers
    else:
        pax_headers = filetar.pax_headers.copy()

    # Parse pax header information. A record looks like that:
    # "%d %s=%s\n" % (length, keyword, value). length is the size
    # of the complete record including the length field itself and
    # the newline. keyword and value are both UTF-8 encoded strings.
    regex = re.compile(r"(\d+) ([^=]+)=", re.U)
    pos = 0
    while True:
        match = regex.match(buf, pos)
        if not match:
            break

        length, keyword = match.groups()
        length = int(length)
        value = buf[match.end(2) + 1:match.start(1) + length - 1]

        try:
            keyword = keyword.decode("utf8")
        except Exception:
            pass

        try:
            value = value.decode("utf8")
        except Exception:
            pass

        pax_headers[keyword] = value
        pos += length

    # Fetch the next header.
    try:
        next = self.fromtarfile(filetar)
    except tarfile.HeaderError:
        raise tarfile.SubsequentHeaderError("missing or bad subsequent header")

    if self.type in (tarfile.XHDTYPE, tarfile.SOLARIS_XHDTYPE):
        # Patch the TarInfo object with the extended header info.
        next._apply_pax_info(pax_headers, filetar.encoding, filetar.errors)
        next.offset = self.offset

        if "size" in pax_headers:
            # If the extended header replaces the size field,
            # we need to recalculate the offset where the next
            # header starts.
            offset = next.offset_data
            if next.isreg() or next.type not in tarfile.SUPPORTED_TYPES:
                offset += next._block(next.size)
            filetar.offset = offset

    return next


def _create_pax_generic_header(cls, pax_headers, type=tarfile.XHDTYPE):
    """Return a POSIX.1-2001 extended or global header sequence
       that contains a list of keyword, value pairs. The values
       must be unicode objects.
    """
    records = []
    for keyword, value in pax_headers.iteritems():

        try:
            keyword = keyword.encode("utf8")
        except Exception:
            pass

        try:
            value = value.encode("utf8")
        except Exception:
            pass

        l = len(keyword) + len(value) + 3   # ' ' + '=' + '\n'
        n = p = 0
        while True:
            n = l + len(str(p))
            if n == p:
                break
            p = n
        records.append("%d %s=%s\n" % (p, keyword, value))
    records = "".join(records)

    # We use a hardcoded "././@PaxHeader" name like star does
    # instead of the one that POSIX recommends.
    info = {}
    info["name"] = "././@PaxHeader"
    info["type"] = type
    info["size"] = len(records)
    info["magic"] = tarfile.POSIX_MAGIC

    # Create pax header + record blocks.
    return cls._create_header(info, tarfile.USTAR_FORMAT) + \
        cls._create_payload(records)

tarfile.TarInfo._proc_pax = _proc_pax
tarfile.TarInfo._create_pax_generic_header = _create_pax_generic_header



#!/usr/bin/python

from setuptools import setup, find_packages
from docker_squash.version import version

import codecs

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name = "docker-squash",
    version = version,
    packages = find_packages(exclude=["tests"]),
    url = 'https://github.com/goldmann/docker-squash',
    download_url = "https://github.com/goldmann/docker-squash/archive/%s.tar.gz" % version,
    author = 'Marek Goldmann',
    author_email = 'marek.goldmann@gmail.com',
    description = 'Docker layer squashing tool',
    license='MIT',
    keywords = 'docker',
    long_description = codecs.open('README.rst', encoding="utf8").read(),
    entry_points = {
        'console_scripts': ['docker-squash=docker_squash.cli:run'],
    },
    tests_require = ['mock'],
    install_requires=requirements
)
